import hashlib
import os
from types import GeneratorType
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Iterable,
    Set,
)

from .base import AbstractSnapshotSerializer


if TYPE_CHECKING:
    from syrupy.types import SerializableData


class DataSerializer:
    indent: str = "  "
    name_marker: str = "# name:"
    divider: str = "---"

    @classmethod
    def write_file(cls, filepath: str, snapshots: Dict[str, Dict[str, Any]]) -> None:
        """
        Writes the snapshot data into the snapshot file that be read later.
        """
        with open(filepath, "w") as f:
            for key in sorted(snapshots.keys()):
                snapshot = snapshots[key]
                snapshot_data = snapshot.get("data")
                if snapshot_data is not None:
                    f.write(f"{cls.name_marker} {key}\n")
                    for data_line in snapshot_data.split("\n"):
                        f.write(f"{cls.indent}{data_line}\n")
                    f.write(f"{cls.divider}\n")

    @classmethod
    def read_file(cls, filepath: str) -> Dict[str, Dict[str, Any]]:
        """
        Read the raw snapshot data (str) from the snapshot file into a dict
        of snapshot name to raw data. This does not attempt any deserialization
        of the snapshot data.
        """
        name_marker_len = len(cls.name_marker)
        indent_len = len(cls.indent)
        snapshots = {}
        test_name = None
        snapshot_data = ""
        try:
            with open(filepath, "r") as f:
                for line in f:
                    if line.startswith(cls.name_marker):
                        test_name = line[name_marker_len:-1].strip(" \n")
                        snapshot_data = ""
                        continue
                    elif test_name is not None:
                        if line.startswith(cls.indent):
                            snapshot_data += line[indent_len:]
                        elif line.startswith(cls.divider) and snapshot_data:
                            snapshots[test_name] = {"data": snapshot_data[:-1]}
        except FileNotFoundError:
            pass

        return snapshots

    @classmethod
    def sort(cls, iterable: Iterable[Any]) -> Iterable[Any]:
        def _sort_key(value: Any) -> Any:
            if isinstance(value, frozenset):
                h = hashlib.sha256()
                for element in cls.sort(value):
                    h.update(str(element).encode("utf-8"))
                return h.hexdigest()
            return value

        try:
            return sorted(iterable)
        except TypeError:
            return sorted(iterable, key=_sort_key)

    @classmethod
    def with_indent(cls, string: str, indent: int) -> str:
        return f"{cls.indent * indent}{string}"

    @classmethod
    def object_type(cls, data: "SerializableData") -> str:
        return f"<class '{data.__class__.__name__}'>"

    @classmethod
    def serialize_string(cls, data: "SerializableData", indent: int = 0) -> str:
        if "\n" in data:
            return (
                cls.with_indent("'\n", indent)
                + str(data)
                + cls.with_indent("\n'", indent)
            )
        return cls.with_indent(repr(data), indent)

    @classmethod
    def serialize_number(cls, data: "SerializableData", indent: int = 0) -> str:
        return cls.with_indent(repr(data), indent)

    @classmethod
    def serialize_set(cls, data: "SerializableData", indent: int = 0) -> str:
        return (
            cls.with_indent(f"{cls.object_type(data)} {{\n", indent)
            + "".join([f"{cls.serialize(d, indent + 1)},\n" for d in cls.sort(data)])
            + cls.with_indent("}", indent)
        )

    @classmethod
    def serialize_dict(cls, data: "SerializableData", indent: int = 0) -> str:
        return (
            cls.with_indent(f"{cls.object_type(data)} {{\n", indent)
            + "".join(
                [
                    (
                        cls.serialize(key, indent + 1)
                        + ": "
                        + cls.serialize(data[key], indent + 1).lstrip(cls.indent)
                        + ",\n"
                    )
                    for key in cls.sort(data.keys())
                ]
            )
            + cls.with_indent("}", indent)
        )

    @classmethod
    def serialize_iterable(cls, data: "SerializableData", indent: int = 0) -> str:
        open_paren, close_paren = next(
            paren[1]
            for paren in {list: "[]", tuple: "()", GeneratorType: "()"}.items()
            if isinstance(data, paren[0])
        )
        return (
            cls.with_indent(f"{cls.object_type(data)} {open_paren}\n", indent)
            + "".join([f"{cls.serialize(d, indent + 1)},\n" for d in data])
            + cls.with_indent(close_paren, indent)
        )

    @classmethod
    def serialize(cls, data: "SerializableData", indent: int = 0) -> str:
        if isinstance(data, str):
            return cls.serialize_string(data, indent)
        elif isinstance(data, (int, float)):
            return cls.serialize_number(data, indent)
        elif isinstance(data, (set, frozenset)):
            return cls.serialize_set(data, indent)
        elif isinstance(data, dict):
            return cls.serialize_dict(data, indent)
        elif isinstance(data, (list, tuple, GeneratorType)):
            return cls.serialize_iterable(data, indent)
        return cls.with_indent(repr(data), indent)


class AmberSnapshotSerializer(AbstractSnapshotSerializer):
    """
    An amber snapshot file stores data in the following format:

    ```
    # name: test_name_1

     data
    ---
    # name: test_name_2

     data
    ```
    """

    @property
    def file_extension(self) -> str:
        return "ambr"

    def discover_snapshots(self, filepath: str) -> Set[str]:
        return set(name for name in DataSerializer.read_file(filepath).keys())

    def _read_snapshot_from_file(
        self, snapshot_file: str, snapshot_name: str
    ) -> "SerializableData":
        snapshots = DataSerializer.read_file(snapshot_file)
        return snapshots.get(snapshot_name, {}).get("data")

    def _write_snapshot_to_file(
        self, snapshot_file: str, snapshot_name: str, data: "SerializableData"
    ) -> None:
        snapshots = DataSerializer.read_file(snapshot_file)
        snapshots[snapshot_name] = {
            "data": self.serialize(data),
        }
        DataSerializer.write_file(snapshot_file, snapshots)

    def delete_snapshots_from_file(
        self, snapshot_file: str, snapshot_names: Set[str]
    ) -> None:
        snapshots = DataSerializer.read_file(snapshot_file)
        for snapshot_name in snapshot_names:
            snapshots.pop(snapshot_name, None)

        if snapshots:
            DataSerializer.write_file(snapshot_file, snapshots)
        else:
            os.remove(snapshot_file)

    def serialize(self, data: "SerializableData") -> str:
        """
        Returns the serialized form of 'data' to be compared
        with the snapshot data written to disk.
        """
        return DataSerializer.serialize(data)



import os
import re
from typing import (
    TYPE_CHECKING,
    Any,
    Set,
)

from .base import AbstractSnapshotSerializer


if TYPE_CHECKING:
    from syrupy.types import SerializableData


class RawSingleSnapshotSerializer(AbstractSnapshotSerializer):
    @property
    def file_extension(self) -> str:
        return "raw"

    def discover_snapshots(self, filepath: str) -> Set[str]:
        """Parse the snapshot name from the filename."""
        return {os.path.splitext(os.path.basename(filepath))[0]}

    def get_file_basename(self, index: int) -> str:
        return self.__clean_filename(self.get_snapshot_name(index=index))

    @property
    def snapshot_subdirectory_name(self) -> str:
        return os.path.splitext(os.path.basename(str(self.test_location.filename)))[0]

    def _read_snapshot_from_file(
        self, snapshot_file: str, snapshot_name: str
    ) -> "SerializableData":
        return self._read_file(snapshot_file)

    def serialize(self, data: "SerializableData") -> bytes:
        return bytes(data)

    def _read_file(self, filepath: str) -> Any:
        try:
            with open(filepath, "rb") as f:
                return f.read()
        except FileNotFoundError:
            return None

    def _write_snapshot_to_file(
        self, snapshot_file: str, snapshot_name: str, data: "SerializableData"
    ) -> None:
        self.__write_file(snapshot_file, data)

    def delete_snapshots_from_file(self, snapshot_file: str, _: Set[str]) -> None:
        os.remove(snapshot_file)

    def __write_file(self, filepath: str, data: "SerializableData") -> None:
        with open(filepath, "wb") as f:
            f.write(self.serialize(data))

    def __clean_filename(self, filename: str) -> str:
        filename = str(filename).strip().replace(" ", "_")
        max_filename_length = 255 - len(self.file_extension or "")
        return re.sub(r"(?u)[^-\w.]", "", filename)[:max_filename_length]



import os

import pytest

from syrupy.utils import walk_snapshot_dir


def makefiles(testdir, filetree, root=""):
    for filename, contents in filetree.items():
        filepath = os.path.join(root, filename)
        if isinstance(contents, dict):
            testdir.mkdir(filepath)
            makefiles(testdir, contents, filepath)
        else:
            name, ext = os.path.splitext(filepath)
            testdir.makefile(ext, **{name: contents})


@pytest.fixture
def testfiles(testdir):
    filetree = {
        "file1.txt": "file1",
        "file2.txt": "file2",
        "__snapshot__": {
            "wrong_snapfile1.ambr": "",
            "wrong_snapfolder": {"wrong_snapfile2.svg": "<svg></svg>"},
        },
        "__snapshots__": {
            "snapfile1.ambr": "",
            "snapfolder": {"snapfile2.svg": "<svg></svg>"},
        },
    }
    makefiles(testdir, filetree)
    return filetree, testdir


def test_walk_dir_skips_non_snapshot_path(testfiles):
    _, testdir = testfiles
    assert {os.path.relpath(p) for p in walk_snapshot_dir(testdir.tmpdir)} == {
        "__snapshots__/snapfile1.ambr",
        "__snapshots__/snapfolder/snapfile2.svg",
    }



from collections import namedtuple

import pytest


def test_non_snapshots(snapshot):
    with pytest.raises(AssertionError):
        assert "Lorem ipsum." == "Muspi merol."


def test_empty_snapshot(snapshot):
    assert snapshot == None  # noqa: E711
    assert snapshot == ""


@pytest.mark.parametrize(
    "actual",
    [
        "",
        r"Raw string",
        r"Escaped \n",
        r"Backslash \u U",
        "🥞🐍🍯",
        "singleline:",
        "- singleline",
        "multi-line\nline 2\nline 3",
        "multi-line\nline 2\n  line 3",
    ],
    ids=lambda x: "",
)
def test_string(snapshot, actual):
    assert snapshot == actual


def test_multiple_snapshots(snapshot):
    assert "First." == snapshot
    snapshot.assert_match("Second.")
    assert snapshot == "Third."


@pytest.mark.parametrize(
    "actual",
    [
        {"b": True, "c": "Some text.", "d": ["1", 2], "a": {"e": False}},
        {"b": True, "c": "Some ttext.", "d": ["1", 2], "a": {"e": False}},
    ],
)
def test_dict(snapshot, actual):
    assert actual == snapshot


def test_set(snapshot):
    assert snapshot == {"this", "is", "a", "set"}
    assert snapshot == {"contains", "frozen", frozenset({"1", "2"})}


ExampleTuple = namedtuple("ExampleTuple", ["a", "b", "c", "d"])


def test_tuple(snapshot):
    assert snapshot == ("this", "is", ("a", "tuple"))
    assert snapshot == ExampleTuple(a="this", b="is", c="a", d={"named", "tuple"})


def test_numbers(snapshot):
    assert snapshot == 3.5
    assert snapshot == 7
    assert snapshot == 2 / 6


def test_list(snapshot):
    assert snapshot == [1, 2, "string", {"key": "value"}]


class TestClass:
    def test_name(self, snapshot):
        assert snapshot == "this is in a test class"



import os

import pytest

from .utils import clean_output


def test_collection(testdir):
    tests = {
        "test_collected": (
            """
            def test_collected(snapshot):
                assert snapshot == 1
                assert snapshot == "2"
                assert snapshot == {1,1,1}
            """
        ),
        "test_not_collected": (
            """
            def test_not_collected(snapshot):
                assert snapshot == "hello"
            """
        ),
    }
    testdir.makepyfile(**tests)
    result = testdir.runpytest("-v", "--snapshot-update")
    result_stdout = clean_output(result.stdout.str())
    assert "4 snapshots generated" in result_stdout

    updated_tests = {
        "test_collected": (
            """
            def test_collected(snapshot):
                assert snapshot == 1
                assert snapshot == "two"
            """
        ),
    }
    testdir.makepyfile(**updated_tests)
    result = testdir.runpytest("-v", "--snapshot-update", "-k", "test_collected")
    result_stdout = clean_output(result.stdout.str())
    assert "1 snapshot passed" in result_stdout
    assert "1 snapshot updated" in result_stdout
    assert "1 snapshot deleted" in result_stdout


def test_collection_parametrized(testdir):
    tests = {
        "test_parametrized": (
            """
            import pytest

            @pytest.mark.parametrize("actual", [1, 2, 3])
            def test_collected(snapshot, actual):
                assert snapshot == actual
            """
        ),
        "test_not_collected": (
            """
            def test_not_collected(snapshot):
                assert snapshot == "hello"
            """
        ),
    }
    testdir.makepyfile(**tests)
    result = testdir.runpytest("-v", "--snapshot-update")
    result_stdout = clean_output(result.stdout.str())
    assert "4 snapshots generated" in result_stdout

    updated_tests = {
        "test_parametrized": (
            """
            import pytest

            @pytest.mark.parametrize("actual", [1, 2])
            def test_collected(snapshot, actual):
                assert snapshot == actual
            """
        ),
    }
    testdir.makepyfile(**updated_tests)
    result = testdir.runpytest("-v", "--snapshot-update", "-k", "test_collected")
    result_stdout = clean_output(result.stdout.str())
    assert "2 snapshots passed" in result_stdout
    assert "snapshot updated" not in result_stdout
    assert "1 snapshot deleted" in result_stdout


@pytest.fixture
def testcases():
    return {
        "inject": (
            """
            def test_injection(snapshot):
                assert snapshot is not None
            """
        ),
        "used": (
            """
            def test_used(snapshot):
                assert snapshot == 'used'
            """
        ),
        "unused": (
            """
            def test_unused(snapshot):
                assert snapshot == 'unused'
            """
        ),
        "updated_1": (
            """
            def test_updated_1(snapshot):
                assert snapshot == ['this', 'will', 'be', 'updated']
            """
        ),
        "updated_2": (
            """
            def test_updated_2(snapshot):
                assert ['this', 'will', 'be', 'updated'] == snapshot
            """
        ),
        "updated_3": (
            """
            def test_updated_3(snapshot):
                assert snapshot == ['this', 'will', 'be', 'updated']
            """
        ),
    }


@pytest.fixture
def testcases_updated(testcases):
    updated_testcases = {
        "updated_1": (
            """
            def test_updated_1(snapshot):
                assert snapshot == ['this', 'will', 'not', 'match']
            """
        ),
        "updated_2": (
            """
            def test_updated_2(snapshot):
                assert ['this', 'will', 'fail'] == snapshot
            """
        ),
        "updated_3": (
            """
            def test_updated_3(snapshot):
                assert snapshot == ['this', 'will', 'be', 'too', 'much']
            """
        ),
    }
    return {**testcases, **updated_testcases}


def test_missing_snapshots(testdir, testcases):
    testdir.makepyfile(test_file=testcases["used"])
    result = testdir.runpytest("-v")
    assert "Snapshot does not exist" in clean_output(result.stdout.str())
    assert result.ret == 1


@pytest.fixture
def stubs(testdir, testcases):
    pyfile_content = "\n\n".join(testcases.values())
    testdir.makepyfile(test_file=pyfile_content)
    filepath = os.path.join(testdir.tmpdir, "__snapshots__", "test_file.ambr")
    return testdir.runpytest("-v", "--snapshot-update"), testdir, testcases, filepath


def test_injected_fixture(stubs):
    result = stubs[0]
    result.stdout.fnmatch_lines(["*::test_injection PASSED*"])
    assert result.ret == 0


def test_generated_snapshots(stubs):
    result = stubs[0]
    result_stdout = clean_output(result.stdout.str())
    assert "5 snapshots generated" in result_stdout
    assert "snapshots unused" not in result_stdout
    assert result.ret == 0


def test_failing_snapshots(stubs, testcases_updated):
    testdir = stubs[1]
    testdir.makepyfile(test_file="\n\n".join(testcases_updated.values()))
    result = testdir.runpytest("-v")
    result_stdout = clean_output(result.stdout.str())
    assert "2 snapshots passed" in result_stdout
    assert "3 snapshots failed" in result_stdout
    expected_strings = [
        # 1
        "-   'be',",
        "+   'not',",
        "-   'updated',",
        "+   'match',",
        # 2
        "-   'be',",
        "+   'fail',",
        "-   'updated',",
        # 3
        "-   'updated',",
        "+   'too',",
        "+   'much',",
    ]
    for string in expected_strings:
        assert string in result_stdout
    assert result.ret == 1


def test_updated_snapshots(stubs, testcases_updated):
    testdir = stubs[1]
    testdir.makepyfile(test_file="\n\n".join(testcases_updated.values()))
    result = testdir.runpytest("-v", "--snapshot-update")
    result_stdout = clean_output(result.stdout.str())
    assert "2 snapshots passed" in result_stdout
    assert "3 snapshots updated" in result_stdout
    assert result.ret == 0


def test_unused_snapshots(stubs):
    _, testdir, tests, _ = stubs
    testdir.makepyfile(test_file="\n\n".join(tests[k] for k in tests if k != "unused"))
    result = testdir.runpytest("-v")
    result_stdout = clean_output(result.stdout.str())
    assert "snapshots generated" not in result_stdout
    assert "4 snapshots passed" in result_stdout
    assert "1 snapshot unused" in result_stdout
    assert result.ret == 0


def test_removed_snapshots(stubs):
    _, testdir, tests, filepath = stubs
    assert os.path.isfile(filepath)
    testdir.makepyfile(test_file="\n\n".join(tests[k] for k in tests if k != "unused"))
    result = testdir.runpytest("-v", "--snapshot-update")
    result_stdout = clean_output(result.stdout.str())
    assert "snapshot unused" not in result_stdout
    assert "1 snapshot deleted" in result_stdout
    assert result.ret == 0
    assert os.path.isfile(filepath)


def test_removed_snapshot_file(stubs):
    _, testdir, tests, filepath = stubs
    assert os.path.isfile(filepath)
    testdir.makepyfile(test_file=tests["inject"])
    result = testdir.runpytest("-v", "--snapshot-update")
    result_stdout = clean_output(result.stdout.str())
    assert "snapshots unused" not in result_stdout
    assert "5 snapshots deleted" in result_stdout
    assert result.ret == 0
    assert not os.path.isfile(filepath)


def test_removed_empty_snapshot_file_only(stubs):
    _, testdir, _, filepath = stubs
    empty_filepath = os.path.join(os.path.dirname(filepath), "test_empty.ambr")
    with open(empty_filepath, "w") as empty_snapfile:
        empty_snapfile.write("")
    assert os.path.isfile(empty_filepath)
    result = testdir.runpytest("-v", "--snapshot-update")
    result_stdout = clean_output(result.stdout.str())
    assert os.path.relpath(filepath) not in result_stdout
    assert "1 snapshot deleted" in result_stdout
    assert "empty snapshot file" in result_stdout
    assert os.path.relpath(empty_filepath) in result_stdout
    assert result.ret == 0
    assert os.path.isfile(filepath)
    assert not os.path.isfile(empty_filepath)


def test_removed_hanging_snapshot_file(stubs):
    _, testdir, _, filepath = stubs
    hanging_filepath = os.path.join(os.path.dirname(filepath), "test_hanging.abc")
    with open(hanging_filepath, "w") as empty_snapfile:
        empty_snapfile.write("some dummy content")
    assert os.path.isfile(hanging_filepath)
    result = testdir.runpytest("-v", "--snapshot-update")
    result_stdout = clean_output(result.stdout.str())
    assert os.path.relpath(filepath) not in result_stdout
    assert "1 snapshot deleted" in result_stdout
    assert "unknown snapshot file" in result_stdout
    assert os.path.relpath(hanging_filepath) in result_stdout
    assert result.ret == 0
    assert os.path.isfile(filepath)
    assert not os.path.isfile(hanging_filepath)



from collections import namedtuple
from itertools import zip_longest
from typing import (
    TYPE_CHECKING,
    Dict,
    List,
    Optional,
    Type,
)

from .constants import SNAPSHOT_EMPTY_FILE
from .exceptions import SnapshotDoesNotExist
from .terminal import (
    error_style,
    green,
    red,
    success_style,
)
from .utils import walk_snapshot_dir


if TYPE_CHECKING:
    from .location import TestLocation
    from .serializers.base import AbstractSnapshotSerializer
    from .session import SnapshotSession
    from .types import SerializableData, SnapshotFiles


AssertionResult = namedtuple(
    "AssertionResult",
    ["asserted", "created", "file", "name", "recalled", "success", "updated"],
)


class SnapshotAssertion:
    def __init__(
        self,
        *,
        update_snapshots: bool,
        serializer_class: Type["AbstractSnapshotSerializer"],
        test_location: "TestLocation",
        session: "SnapshotSession",
    ):
        self._update_snapshots = update_snapshots
        self._serializer_class = serializer_class
        self._test_location = test_location
        self._executions = 0
        self._execution_results: Dict[int, "AssertionResult"] = {}

        self._session: "SnapshotSession" = session
        self._session.register_request(self)

    @property
    def serializer(self) -> "AbstractSnapshotSerializer":
        if not getattr(self, "_serializer", None):
            self._serializer: "AbstractSnapshotSerializer" = self._serializer_class(
                test_location=self._test_location
            )
        return self._serializer

    @property
    def num_executions(self) -> int:
        return int(self._executions)

    @property
    def executions(self) -> Dict[int, AssertionResult]:
        return self._execution_results

    @property
    def discovered_snapshots(self) -> "SnapshotFiles":
        return {
            filepath: (
                self.serializer.discover_snapshots(filepath) or SNAPSHOT_EMPTY_FILE
            )
            if filepath.endswith(self.serializer.file_extension)
            else set()
            for filepath in walk_snapshot_dir(self.serializer.dirname)
        }

    def with_class(
        self, serializer_class: Optional[Type["AbstractSnapshotSerializer"]] = None,
    ) -> "SnapshotAssertion":
        return self.__class__(
            update_snapshots=self._update_snapshots,
            test_location=self._test_location,
            serializer_class=serializer_class or self._serializer_class,
            session=self._session,
        )

    def assert_match(self, data: "SerializableData") -> None:
        assert self == data

    def get_assert_diff(self, data: "SerializableData") -> List[str]:
        assertion_result = self._execution_results[self.num_executions - 1]
        snapshot_data = assertion_result.recalled
        serialized_data = self.serializer.serialize(data)
        if snapshot_data is None:
            return ["Snapshot does not exist!"]

        diff = []
        if not assertion_result.success:
            received = serialized_data.splitlines()
            stored = snapshot_data.splitlines()

            marker_stored = success_style("-")
            marker_received = error_style("+")

            for received_line, stored_line in zip_longest(received, stored):
                if received_line is None:
                    diff.append(f"{marker_stored} {green(stored_line)}")
                elif stored_line is None:
                    diff.append(f"{marker_received} {red(received_line)}")
                elif received_line != stored_line:
                    diff.extend(
                        [
                            f"{marker_stored} {green(stored_line)}",
                            f"{marker_received} {red(received_line)}",
                        ]
                    )

        return diff

    def __repr__(self) -> str:
        return f"<SnapshotAssertion ({self.num_executions})>"

    def __eq__(self, other: "SerializableData") -> bool:
        return self._assert(other)

    def _assert(self, data: "SerializableData") -> bool:
        matches = False
        assertion_success = False
        snapshot_data = None
        serialized_data = None
        try:
            snapshot_file = self.serializer.get_filepath(self.num_executions)
            snapshot_name = self.serializer.get_snapshot_name(self.num_executions)
            snapshot_data = self._recall_data(index=self.num_executions)
            serialized_data = self.serializer.serialize(data)
            matches = snapshot_data is not None and serialized_data == snapshot_data
            assertion_success = matches
            if not matches and self._update_snapshots:
                self.serializer.create_or_update_snapshot(
                    data=data, index=self.num_executions
                )
                assertion_success = True
            return assertion_success
        finally:
            snapshot_created = snapshot_data is None and assertion_success
            snapshot_updated = matches is False and assertion_success
            self._execution_results[self._executions] = AssertionResult(
                file=snapshot_file,
                name=snapshot_name,
                recalled=snapshot_data,
                asserted=serialized_data,
                success=assertion_success,
                created=snapshot_created,
                updated=snapshot_updated,
            )
            self._executions += 1

    def _recall_data(self, index: int) -> Optional["SerializableData"]:
        try:
            return self.serializer.read_snapshot(index=index)
        except SnapshotDoesNotExist:
            return None



from typing import TYPE_CHECKING

from .raw_single import RawSingleSnapshotSerializer


if TYPE_CHECKING:
    from syrupy.types import SerializableData


class PNGImageSnapshotSerializer(RawSingleSnapshotSerializer):
    @property
    def file_extension(self) -> str:
        return "png"

    def serialize(self, data: "SerializableData") -> bytes:
        if isinstance(data, bytes):
            return data
        raise TypeError(
            "Can not serialize image data. Expected "
            f"'bytes', got '{type(data).__name__}'."
        )


class SVGImageSnapshotSerializer(RawSingleSnapshotSerializer):
    @property
    def file_extension(self) -> str:
        return "svg"

    def serialize(self, data: "SerializableData") -> bytes:
        return str(data).encode("utf-8")



import inspect
from typing import (
    Any,
    Optional,
)


class TestLocation(object):
    def __init__(self, node: Any):
        self._node = node
        self.filename = self._node.fspath
        self.modulename = self._node.obj.__module__
        self.methodname = self._node.obj.__name__
        self.nodename = getattr(self._node, "name", None)
        self.testname = self.nodename or self.methodname

    @property
    def classname(self) -> Optional[str]:
        if inspect.ismethod(self._node.obj):
            method = self._node.obj
            for cls in inspect.getmro(method.__self__.__class__):
                if method.__name__ in cls.__dict__:
                    return cls.__name__
        return None

    def matches_snapshot_name(self, snapshot_name: str) -> bool:
        matches_basemethod = str(self.methodname) in snapshot_name
        matches_testnode = snapshot_name in str(self.nodename)
        return matches_basemethod or matches_testnode



import pytest

from syrupy.serializers.amber import DataSerializer


@pytest.mark.parametrize(
    "data",
    ["some string", "string with 'quotes'", r"Raw string", ["a", "list", 1, 2]],
    ids=lambda x: "",
)
def test_does_not_fail(data):
    DataSerializer.serialize(data)


class SomeClass:
    pass


@pytest.mark.parametrize(
    "data",
    ["a string", 123, (1, 2), True, None, dict(), frozenset(), SomeClass()],
    ids=lambda x: "",
)
def test_produces_valid_object_types(snapshot, data):
    assert snapshot == DataSerializer.object_type(data)
